

#install.packages("matrixStats")

library(matrixStats)

#install.packages("numDeriv")
library(numDeriv)
library(pracma)
#library()

## Consider a gaussian with mean = 1, variance = 0.2. 
## Fit (mu, sigma) using maximum likelihood, L2, and ICE. Also consider the version where we fit 1/sigma instad of sigma. 

buckets <- 1000.0


probCalculation <- function(X, theta) {
	objective_mean <- theta[1]
	objective_sigma <- theta[2]
	objective_var <- objective_sigma * objective_sigma
	centered <- X - objective_mean
	exponent <- (centered * centered) / (2 * objective_var)
	rawProb <- exp(-exponent) / sqrt(2 * pi * objective_var)
	rawProb <- rawProb / buckets
	rawProb
}

computeEntropy <- function(X, theta) {
	objective_mean <- theta[1]
	objective_sigma <- theta[2]
	objective_var <- objective_sigma * objective_sigma
	centered <- X - objective_mean
	exponent <- (centered * centered) / (2 * objective_var)
	
	entropy <- exponent + 0.5*log(2 * pi * objective_var) + log(buckets)
	return (entropy)
}

objective_mle <- function(x, theta, l1, l2) {
	entropy <- computeEntropy(x, theta)	
	mean_entropy = mean(entropy)
	
	l1Norm <- sum(abs(theta))
	l2Norm <- sum(theta * theta)
	
	result <- mean_entropy + (l1 * l1Norm) + (l2 * l2Norm)
	
	return (result)
}

objective_ice <- function(x, theta) {
	return (objective_ice_raw(x, theta, FALSE))
}

objective_ice_grad <- function(x, theta) {
	return (objective_ice_raw(x, theta, TRUE))
}

ice_matrix <- function(x, theta) {
	objective_mean <- theta[1]
	objective_sigma <- abs(theta[2]) # no-op since sigma always enters squared. 
	
	z <- (x - objective_mean) / objective_sigma
	
	d1 <- (-1 / objective_sigma) * matrix(c((z), (z*z - 1)), nrow=length(z), ncol=2)
	ez = mean(z)
	ez2 = mean(z*z)
	ez3 = mean(z*z*z)
	ez4 = mean(z*z*z*z)
	s2 = objective_sigma* objective_sigma

	iMatrix <- (1 / length(z)) * (t(d1) %*% d1)
	jMatrix <- (1 / s2) * matrix(c(1, 2*ez, 2*ez, 3*ez2 - 1), nrow=2, ncol=2)
	output_matrix = iMatrix %*% pinv(jMatrix)
	return (output_matrix)
}

ice_wrap <- function(X, x) {
	return (objective_mle(X, x, 0.0, 0.0))
}

ice_iMatrix <- function(x, theta) {
	i_mat = c(0, 0) %*% t(c(0,0))
	
	for(i in 1:length(x)) {
		j_vec = jacobian(f=ice_wrap, x=theta, X=x[i])
		i_mat = i_mat + t(j_vec) %*% j_vec		
	}
	
	return (i_mat) / length(x)
}

ice_jMatrix <- function(x, theta) {
	objective_mean <- theta[1]
	objective_sigma <- abs(theta[2]) # no-op since sigma always enters squared. 
	
	z <- (x - objective_mean) / objective_sigma
	
	d1 <- (-1 / objective_sigma) * matrix(c((z), (z*z - 1)), nrow=length(z), ncol=2)
	ez = mean(z)
	ez2 = mean(z*z)
	s2 = objective_sigma* objective_sigma
	
	jMatrix <- (1 / s2) * matrix(c(1, 2*ez, 2*ez, 3*ez2 - 1), nrow=2, ncol=2)
	return (jMatrix)
}


objective_ice_raw <- function(x, theta, do_grad) {
	entropy <- computeEntropy(x, theta)	
	mean_entropy = mean(entropy)
	
	objective_mean <- theta[1]
	objective_sigma <- abs(theta[2]) # no-op since sigma always enters squared. 
	
	z <- (x - objective_mean) / objective_sigma
	
	d1 <- (-1 / objective_sigma) * matrix(c((z), (z*z - 1)), nrow=length(z), ncol=2)
	ez = mean(z)
	ez2 = mean(z*z)
	ez3 = mean(z*z*z)
	ez4 = mean(z*z*z*z)
	s2 = objective_sigma* objective_sigma
	
	# fairly nasty, but basically compute individual I matrices, then average them. 
	#iMatrix <- (1 / length(z)) * (t(d1) %*% d1)
	# Or compute symbolically. 
	#iMatrix <- (1 / s2) * matrix(c(ez2, ez3 - ez, ez3 - ez, ez4 - 2*ez2 + 1), nrow=2, ncol=2) 
	
	#jMatrix <- (1 / s2) * matrix(c(1, 2*ez, 2*ez, 3*ez2 - 1), nrow=2, ncol=2)
	#jInverse <- solve(jMatrix)
	
	# j2 = -hessian(f=hf, x=theta)
	# i2 = jacobian(f=hf, x=theta)
	# i2 = t(i2) %*% i2
	# j2i = pinv(j2)
	
	# We can compute the inverse symbolically. 
	d <- 1 / (3*ez2 - 1 - 4*ez*ez) # a common value
	#jInverse <- s2 * d * matrix(c((3*ez2 - 1), -2*ez, -2*ez, 1), nrow=2, ncol=2)
	
	iceMatrix = ice_matrix(x, theta)
	iceCorrection2 = max(sum(diag(iceMatrix)), 0) / length(x)
	#trace = sum(diag(iMatrix %*% jInverse))
	#iceCorrection = max((trace / length(x)), 0)
	
	

	# We can show that trace(m) = trace(m2)
	# where m2 was computed symbolically from symbolic jInverse and I. 
	#m <- iMatrix %*% jInverse
	#m2 <- d * matrix(c((3*ez2*ez2 - ez2 -  2*ez*ez3 + 2*ez*ez), 0, 0, -2*ez*ez3 + 2*ez*ez + ez4 - 2*ez2 + 1), nrow=2, ncol=2)
	
	# This is the trace of I*JInverse
	numerator = ((3*ez2*ez2) - (3*ez2) - (4*ez*ez3) + ez4 + (4*ez*ez) + 1)
	t2 = d * numerator
	iceCorrection = max(t2, 0) / length(x)
	
	#iceCorrection2 = max(sum(diag(ice_matrix(x, c(0.2, 0.2)))), 0) / length(x)
	#iceCorrection2 = max(sum(diag(iMatrix %*% pinv(j2))), 0) / length(x)
	#print(paste("Correction diff[", iceCorrection, "]: ", (iceCorrection2 - iceCorrection)))
	iceCorrection = iceCorrection2
	
	
	
	if(do_grad) {
		grad_ice = c(0.0, 0.0)
		
		if(t2 > 0) {
			grad_a_mu = d*(12*ez2*ez - 6*ez - 4 *ez3 -12 * ez * ez2 + 4 * ez3 + 8*ez)
			grad_b_mu = numerator * d * d * (6*ez - 8*ez)
			grad_core_mu = grad_a_mu - grad_b_mu
			
			grad_a_sigma = d*(12*ez2*ez2 - 6*ez2 - 4 *ez * ez3 - 12 * ez * ez3 + 4 * ez4 + 8*ez*ez)
			grad_b_sigma = numerator * d * d * (6*ez2 - 8*ez*ez)
			grad_core_sigma = grad_a_sigma - grad_b_sigma
			
			grad_mu = -grad_core_mu / objective_sigma
			grad_sigma = -grad_core_sigma / objective_sigma
			grad_ice = c(grad_mu, grad_sigma)
			grad_ice = grad_ice / length(x)
		}
		
		grad_entropy = colMeans(d1)
		total_grad = grad_entropy + grad_ice
		
		#print(paste("Total grad: ", total_grad[1], ", ", total_grad[2]))
	
		return (total_grad)
	}
	
	# Check to see that we get the expected entropy.
	# expected <- 0.5 * (z * z) + log(buckets) + 0.5*log(2 * pi * objective_sigma * objective_sigma)

	result <- mean_entropy + iceCorrection
	return (result)
}

#jacobian(f=ice_wrap, x=theta, X=x)
#hessian(f=ice_wrap, x=theta, X=x)
#jacobian(f=objective_ice_raw, x=theta, X=x)

solve_mle <- function(X, startTheta, L1, L2) {
	if(L2 != 0 && L1 != 0) {
		startTheta = solve_mle(X, startTheta, 0.0, 0.0)
	}
	
	mle_val <- optim(startTheta, objective_mle, gr = NULL, x = X, l1 = L1, l2 = L2, control=list(maxit=10000))
	return (mle_val$par)
}


solve_ice <- function(X, startTheta) {
	# We must always start from the MLE estimate.
	starting_point = solve_mle(X, startTheta, 0.0, 0.0)
	
	mle_val <- optim(starting_point, objective_ice, method="BFGS", gr = NULL, x = X, control=list(maxit=10000))
	
	best_val = mle_val$val
	best_par = mle_val$par
	
	for(i in 1:1) {
		#adj = 0.01*rnorm(2)
		mle_val <- optim(starting_point, objective_ice, method="BFGS", gr = NULL, x = X, control=list(maxit=10000))
		#mle_val <- optim(starting_point + adj, objective_ice, method="BFGS", gr = NULL, x = X, control=list(maxit=10000))
		#return (mle_val$par)
		
		if(mle_val$val < best_val) {
			#print(paste("New best: ", best_val, " -> ", mle_val$val))
			best_val = mle_val$val
			best_par = mle_val$par
		}
		else 
		{
			#print(paste("Unimproved: ", best_val, " < ", mle_val$val))
		}
	}
	
	return (best_par)
}

snapToGrid <- function(inputVal, bucketCount) {
	return (floor(inputVal * bucketCount) / bucketCount)
}


compute_bucketset <- function(theta) {
	mu <- theta[1]
	sigma <- theta[2]
	
	lower <- mu - (10 * sigma);
	upper <- mu + (10 * sigma);
	intervalSize <- upper - lower
	bucketSet <- lower + (0:floor(intervalSize * buckets) / buckets)
	
	return (bucketSet)
}


rho <- function(theta0, theta) {
	bucketSet <- compute_bucketset(theta0)
	
	p <- probCalculation(bucketSet, theta0)
	q <- probCalculation(bucketSet, theta)
	
	ratio <- pmax(p / q, 1.0e-20) # floor this out, lim p -> 0 of p log(p) = 0. 
	kl <- sum(p * log(ratio))
	return (kl)
}

hf <- function(theta0) {
	bucketSet <- compute_bucketset(theta0)	
	p <- pmax(probCalculation(bucketSet, theta0), 1.0e-20)
	hval <- -sum(p*log(p))
	return (hval)
}


run_test <- function(mu, sigma, startTheta, n, iteration) {
	#Generate test data.
	# LDDP this thing with 5 sigma cutoffs. 
	# snap mean and sigma to grid. 
	mu <- snapToGrid(mu, buckets)
	sigma <- max(snapToGrid(sigma, buckets), 0.1)
	
	bucketSet <- compute_bucketset(c(mu, sigma))
	#print(paste("Bucket sum: ", sum(probCalculation(bucketSet, c(mu, sigma))))) # Should be extremely close to 1.0
	
	X <- (sigma * rnorm(n)) + mu
	X <- snapToGrid(X, buckets)
	
	np = 20 * 1000
	Xp <- (sigma * rnorm(np)) + mu
	Xp <- snapToGrid(Xp, buckets)
	#assign("Xp_g", Xp, envir = .GlobalEnv)
	
	mle <- solve_mle(X, startTheta, 0.0, 0.0)
	mle_entropy <- objective_mle(X, mle, 0.0, 0.0)
	mle_entropy_adj <- objective_ice_raw(X, mle, FALSE)

	ice_theta = solve_ice(X, startTheta)
	ice_entropy <- objective_mle(X, ice_theta, 0.0, 0.0)
	ice_entropy_adj <- objective_ice_raw(X, ice_theta, FALSE)
	
	true_entropy <- objective_mle(X, c(mu, sigma), 0.0, 0.0)
	hf <- hf(c(mu, sigma))

	mle_kl = rho(c(mu, sigma), mle)
	ice_kl = rho(c(mu, sigma), ice_theta)

	ice_test = objective_mle(Xp, ice_theta, 0.0, 0.0)
	mle_test = objective_mle(Xp, mle, 0.0, 0.0)

	iMatrix = ice_iMatrix(X, ice_theta)
	jMatrix = ice_jMatrix(X, ice_theta)

	min_condition = min(rcond(iMatrix), rcond(jMatrix))
	
	a = pinv(jMatrix)
	b = pinv(diag(2)*jMatrix)
	
	iceMatrix1 = iMatrix %*% a
	iceMatrix2 = iMatrix %*% b
	iceCorrection1 = max(sum(diag(iceMatrix1)), 0) / length(X)
	iceCorrection2 = max(sum(diag(iceMatrix2)), 0) / length(X)

	norm1 = norm(a)
	norm2 = norm(b)
	norm3 = norm(a - b)	
	
	print(paste(n, ", ", iteration, ", ", mle_kl, ", ", ice_kl, ", ", mle_entropy, ", ", ice_entropy, ", ", mle_entropy_adj, ", ", ice_entropy_adj, ", ", hf, ", ", mle_test, ", ", ice_test, ", ", min_condition, ", ", iceCorrection1, ", ", iceCorrection2, ", ", norm1, ", ", norm2, ", ", norm3))

	
	
	# print(paste(n, ", ", iteration, ", ", mu, ", ", sigma, ", ", mle[1], ", ", mle[2], ", ", lambda2, ", ", l2_theta[1], ", ", l2_theta[2], ", ", l2a_theta[1], ", ", l2a_theta[2], ", ", ice_theta[1], ", ", ice_theta[2], ", ", mle_kl, ", ", l2_kl, ", ", l2a_kl, ", ", ice_kl, ", ", mle_entropy, ", ", l2_entropy, ", ", l2a_entropy, ", ", ice_entropy, ", ", true_entropy, ", ", hf))
	
	
	return (c(mle_kl, ice_kl, mle_entropy, ice_entropy, true_entropy, hf, mle_test, ice_test))
}

run_test_repeated <- function(mu, sigma, startTheta, n, count) {
	results <- c()

	print("n, iteration, mle_kl, ice_kl, mle_entropy, ice_entropy, mle_entropy_adj, ice_entropy_adj, hf, mle_test, ice_test, min_condition, jCorrecetion, dCorrection, jNorm, dNorm, diffNorm")
	
	for(k in 1:count) {
		output <- run_test(mu, sigma, startTheta, n, k)
		results <- cbind(c(output), results)
	}
	
	repeated_results <- colMeans(t(results))
	repeated_devs <- sqrt(colVars(t(results)) / n)
	ice_diffs = (t(results[2,]) - t(results[1,]))[1,]
	ice_dev = sqrt(var(ice_diffs) / count)
	ice_z = mean(ice_diffs) / ice_dev;
	
	
	print(paste("MLE KL: ", repeated_results[1]))
	print(paste("ICE KL: ", repeated_results[2], " (worse (or better if negative) by ", (repeated_results[2] - repeated_results[1]), ") [", ice_z , "]"))
	
	
	print(paste("Summary: ", repeated_results[1], ", ", repeated_devs[1], ", ", repeated_results[2], ", ", repeated_devs[2], ", ", ice_z))
	
	# mle_train = repeated_results[3]
	# ice_train = repeated_results[4]
	# true_entropy = repeated_results[5]
	# hf = repeated_results[6]
	# mle_test = repeated_results[1] + hf
	# ice_test = repeated_results[2] + hf
	# mle_test2 = repeated_results[7]
	# ice_test2 = repeated_results[8]
	
	# ice_bias = ice_test - ice_train
	# mle_bias = mle_test - mle_train
	
	# ice_bias_norm_n = ice_bias * n
	# mle_bias_norm_n = mle_bias * n
	# ice_bias_norm_n32 = ice_bias * n * sqrt(n)
	# mle_bias_norm_n32 = mle_bias * n * sqrt(n)
	
	# print(paste("Bias comparison[MLE]: ", mle_bias, ", ", mle_bias_norm_n, ", ", mle_bias_norm_n32))
	# print(paste("Bias comparison[ICE]: ", ice_bias, ", ", ice_bias_norm_n, ", ", ice_bias_norm_n32))
	
	# ice_bias = ice_test2 - ice_train
	# mle_bias = mle_test2 - mle_train
	
	# ice_bias_norm_n = ice_bias * n
	# mle_bias_norm_n = mle_bias * n
	# ice_bias_norm_n32 = ice_bias * n * sqrt(n)
	# mle_bias_norm_n32 = mle_bias * n * sqrt(n)
	
	# print(paste("Bias comparison[MLE2]: ", mle_bias, ", ", mle_bias_norm_n, ", ", mle_bias_norm_n32))
	# print(paste("Bias comparison[ICE2]: ", ice_bias, ", ", ice_bias_norm_n, ", ", ice_bias_norm_n32))
	
	return (repeated_results)
}


meanTest = rnorm(1)[1]
sigmaTest = abs(rnorm(1)[1])

start <- c(1.0, 1.0)

run_test(0.0, 0.2, start, 10, 0)


# This should almost always show that L2 is worse than MLE. 
# repeated_results <- run_test_repeated(0.2, 0.2, start, 4, 500)
repeated_results <- run_test_repeated(0.2, 0.2, start, 8, 2000)
repeated_results <- run_test_repeated(0.2, 0.2, start, 16, 2000)
repeated_results <- run_test_repeated(0.2, 0.2, start, 32, 2000)
repeated_results <- run_test_repeated(0.2, 0.2, start, 64, 2000)
repeated_results <- run_test_repeated(0.2, 0.2, start, 128, 2000)
repeated_results <- run_test_repeated(0.2, 0.2, start, 256, 2000)
repeated_results <- run_test_repeated(0.2, 0.2, start, 512, 2000)
repeated_results <- run_test_repeated(0.2, 0.2, start, 1024, 2000)

# print(paste("MLE KL: ", repeated_results[1]))
# print(paste("L2 KL: ", repeated_results[2], " (worse by ", (repeated_results[2] - repeated_results[1]), ")"))
# print(paste("L2a KL: ", repeated_results[3], " (worse by ", (repeated_results[3] - repeated_results[1]), ")"))
# print(paste("ICE KL: ", repeated_results[4], " (worse (or better if negative) by ", (repeated_results[4] - repeated_results[1]), ")"))

