

#install.packages("matrixStats")

library(matrixStats)
#library(numDeriv)
library(pracma)

## Consider a gaussian with mean = 1, variance = 0.2. 
## Fit (mu, sigma) using maximum likelihood, L2, and ICE. Also consider the version where we fit 1/sigma instad of sigma. 

buckets <- 1000.0


probCalculation <- function(X, theta) {
	objective_mean <- theta[1]
	objective_sigma <- theta[2]
	objective_var <- objective_sigma * objective_sigma
	centered <- X - objective_mean
	exponent <- (centered * centered) / (2 * objective_var)
	rawProb <- exp(-exponent) / sqrt(2 * pi * objective_var)
	rawProb <- rawProb / buckets
	rawProb
}

computeEntropy <- function(X, theta) {
	objective_mean <- theta[1]
	objective_sigma <- theta[2]
	objective_var <- objective_sigma * objective_sigma
	centered <- X - objective_mean
	exponent <- (centered * centered) / (2 * objective_var)
	
	entropy <- exponent + 0.5*log(2 * pi * objective_var) + log(buckets)
	return (entropy)
}

objective_mle <- function(x, theta, l1, l2) {
	entropy <- computeEntropy(x, theta)	
	mean_entropy = mean(entropy)
	
	l1Norm <- sum(abs(theta))
	l2Norm <- sum(theta * theta)
	
	result <- mean_entropy + (l1 * l1Norm) + (l2 * l2Norm)
	
	return (result)
}

invert_sigma <- function(theta) {
	thetaPrime <- theta
	thetaPrime[2] <- 1.0 / thetaPrime[2]
	return (thetaPrime)
}

#Same as above, but using 1/sigma instead of sigma.
objective_mle2 <- function(x, theta, l1, l2) {
	thetaPrime <- invert_sigma(theta)
	entropy <- computeEntropy(x, thetaPrime)	
	mean_entropy = mean(entropy)
	
	l1Norm <- sum(abs(thetaPrime))
	l2Norm <- sum(thetaPrime * thetaPrime)
	
	result <- mean_entropy + l1 * l1Norm + l2 * l2Norm
	
	return (result)
}


ice_matrix <- function(x, theta) {
	objective_mean <- theta[1]
	objective_sigma <- abs(theta[2]) # no-op since sigma always enters squared. 
	
	z <- (x - objective_mean) / objective_sigma
	
	d1 <- (-1 / objective_sigma) * matrix(c((z), (z*z - 1)), nrow=length(z), ncol=2)
	ez = mean(z)
	ez2 = mean(z*z)
	ez3 = mean(z*z*z)
	ez4 = mean(z*z*z*z)
	s2 = objective_sigma* objective_sigma

	iMatrix <- (1 / length(z)) * (t(d1) %*% d1)
	jMatrix <- (1 / s2) * matrix(c(1, 2*ez, 2*ez, 3*ez2 - 1), nrow=2, ncol=2)
	output_matrix = iMatrix %*% pinv(jMatrix)
	return (output_matrix)
}

ice_wrap <- function(X, x) {
	return (objective_mle(X, x, 0.0, 0.0))
}

ice_iMatrix <- function(w, theta) {
	i_mat = c(0, 0) %*% t(c(0,0))
	
	for(i in 1:length(w)) {
		j_vec = jacobian(f=ice_wrap, x=theta, X=w[i])
		i_mat = i_mat + t(j_vec) %*% j_vec		
	}
	
	return (i_mat / length(w))
}

ice_jMatrix <- function(x, theta) {
	objective_mean <- theta[1]
	objective_sigma <- abs(theta[2]) # no-op since sigma always enters squared. 
	
	z <- (x - objective_mean) / objective_sigma
	
	d1 <- (-1 / objective_sigma) * matrix(c((z), (z*z - 1)), nrow=length(z), ncol=2)
	ez = mean(z)
	ez2 = mean(z*z)
	s2 = objective_sigma* objective_sigma
	
	jMatrix <- (1 / s2) * matrix(c(1, 2*ez, 2*ez, 3*ez2 - 1), nrow=2, ncol=2)
	return (jMatrix)
}


objective_ice_raw <- function(x, xp, theta, do_grad) {
	entropy <- computeEntropy(x, theta)	
	mean_entropy = mean(entropy)
	
	objective_mean <- theta[1]
	objective_sigma <- abs(theta[2]) # no-op since sigma always enters squared. 
	
	z <- (xp - objective_mean) / objective_sigma
	
	d1 <- (-1 / objective_sigma) * matrix(c((z), (z*z - 1)), nrow=length(z), ncol=2)
	ez = mean(z)
	ez2 = mean(z*z)
	ez3 = mean(z*z*z)
	ez4 = mean(z*z*z*z)
	s2 = objective_sigma* objective_sigma
	
	# fairly nasty, but basically compute individual I matrices, then average them. 
	#iMatrix <- (1 / length(z)) * (t(d1) %*% d1)
	# Or compute symbolically. 
	#iMatrix <- (1 / s2) * matrix(c(ez2, ez3 - ez, ez3 - ez, ez4 - 2*ez2 + 1), nrow=2, ncol=2) 
	
	#jMatrix <- (1 / s2) * matrix(c(1, 2*ez, 2*ez, 3*ez2 - 1), nrow=2, ncol=2)
	#jInverse <- solve(jMatrix)
	
	# j2 = -hessian(f=hf, x=theta)
	# i2 = jacobian(f=hf, x=theta)
	# i2 = t(i2) %*% i2
	# j2i = pinv(j2)
	
	# We can compute the inverse symbolically. 
	d <- 1 / (3*ez2 - 1 - 4*ez*ez) # a common value
	#jInverse <- s2 * d * matrix(c((3*ez2 - 1), -2*ez, -2*ez, 1), nrow=2, ncol=2)
	
	iceMatrix = ice_matrix(x, theta)
	iceCorrection2 = max(sum(diag(iceMatrix)), 0) / length(x)
	#trace = sum(diag(iMatrix %*% jInverse))
	#iceCorrection = max((trace / length(x)), 0)
	
	

	# We can show that trace(m) = trace(m2)
	# where m2 was computed symbolically from symbolic jInverse and I. 
	#m <- iMatrix %*% jInverse
	#m2 <- d * matrix(c((3*ez2*ez2 - ez2 -  2*ez*ez3 + 2*ez*ez), 0, 0, -2*ez*ez3 + 2*ez*ez + ez4 - 2*ez2 + 1), nrow=2, ncol=2)
	
	# This is the trace of I*JInverse
	#numerator = ((3*ez2*ez2) - (3*ez2) - (4*ez*ez3) + ez4 + (4*ez*ez) + 1)
	#t2 = d * numerator
	#iceCorrection = max(t2, 0) / length(x)
	
	#iceCorrection2 = max(sum(diag(ice_matrix(x, c(0.2, 0.2)))), 0) / length(x)
	#iceCorrection2 = max(sum(diag(iMatrix %*% pinv(j2))), 0) / length(x)
	#print(paste("Correction diff[", iceCorrection, "]: ", (iceCorrection2 - iceCorrection)))
	iceCorrection = iceCorrection2
	
	
	
	if(do_grad) {
		grad_ice = c(0.0, 0.0)
		
		if(t2 > 0) {
			grad_a_mu = d*(12*ez2*ez - 6*ez - 4 *ez3 -12 * ez * ez2 + 4 * ez3 + 8*ez)
			grad_b_mu = numerator * d * d * (6*ez - 8*ez)
			grad_core_mu = grad_a_mu - grad_b_mu
			
			grad_a_sigma = d*(12*ez2*ez2 - 6*ez2 - 4 *ez * ez3 - 12 * ez * ez3 + 4 * ez4 + 8*ez*ez)
			grad_b_sigma = numerator * d * d * (6*ez2 - 8*ez*ez)
			grad_core_sigma = grad_a_sigma - grad_b_sigma
			
			grad_mu = -grad_core_mu / objective_sigma
			grad_sigma = -grad_core_sigma / objective_sigma
			grad_ice = c(grad_mu, grad_sigma)
			grad_ice = grad_ice / length(x)
		}
		
		grad_entropy = colMeans(d1)
		total_grad = grad_entropy + grad_ice
		
		#print(paste("Total grad: ", total_grad[1], ", ", total_grad[2]))
	
		return (total_grad)
	}
	
	# Check to see that we get the expected entropy.
	# expected <- 0.5 * (z * z) + log(buckets) + 0.5*log(2 * pi * objective_sigma * objective_sigma)

	result <- mean_entropy + iceCorrection
	return (result)
}

objective_ice <- function(x, xp, theta) {
	return (objective_ice_raw(x, xp, theta, FALSE))
}

objective_ice_grad <- function(x, xp, theta) {
	return (objective_ice_raw(x, xp, theta, TRUE))
}


solve_mle <- function(X, startTheta, L1, L2) {
	if(L2 != 0 && L1 != 0) {
		startTheta = solve_mle(X, startTheta, 0.0, 0.0)
	}
	
	mle_val <- optim(startTheta, objective_mle, gr = NULL, x = X, l1 = L1, l2 = L2, control=list(maxit=10000))
	return (mle_val$par)
}

solve_mle2 <- function(X, startTheta, L1, L2) {
	if(L2 != 0 && L1 != 0) {
		startTheta = solve_mle(X, startTheta, 0.0, 0.0)
	}
	
	startThetaPrime <- invert_sigma(startTheta)
	
	mle_val <- optim(startThetaPrime, objective_mle2, gr = NULL, x = X, l1 = L1, l2 = L2, control=list(maxit=10000))

	# switch back to the traditional definition.
	thetaSolved <- invert_sigma(mle_val$par)

	return (thetaSolved)
}

solve_ice <- function(X, Xp, startTheta) {
	# We must always start from the MLE estimate.
	starting_point = solve_mle(X, startTheta, 0.0, 0.0)

	mle_val <- optim(starting_point, objective_ice, gr = NULL, x = X, xp = Xp, control=list(maxit=10000))
	return (mle_val$par)
}

snapToGrid <- function(inputVal, bucketCount) {
	return (floor(inputVal * bucketCount) / bucketCount)
}

lambda_crossval_objective <- function(X, start, lambda1, lambda2) {
	oos_total <- 0.0
	n_splits <- min(length(X)/2, 4)
	
	for(k in 1:n_splits) {
		test <- seq(k, length(X), n_splits)
		train <- seq(1, length(X), 1)
		train <- train[-test]
		
		X_train <- X[train]
		X_test <- X[test]
		
		nextTheta <- solve_mle(X_train, start, lambda1, lambda2)
		oos_next <- objective_mle(X_test, nextTheta, lambda1, lambda2)
		oos_total <- oos_total + oos_next
	}
	
	return (oos_total)
}

#Same, but use the inverted sigma definition. 
lambda_crossval_objective2 <- function(X, start, lambda1, lambda2) {
	oos_total <- 0.0
	n_splits <- min(length(X)/2, 4)
	
	for(k in 1:n_splits) {
		test <- seq(k, length(X), n_splits)
		train <- seq(1, length(X), 1)
		train <- train[-test]
		
		X_train <- X[train]
		X_test <- X[test]
		
		nextTheta <- solve_mle2(X_train, start, lambda1, lambda2)
		oos_next <- objective_mle(X_test, nextTheta, lambda1, lambda2)
		oos_total <- oos_total + oos_next
	}
	
	return (oos_total)
}


lambda_crossval_ridge <- function(X, start, lambda) {
	return (lambda_crossval_objective(X, start, lambda1 = 0.0, lambda))
}

lambda_crossval_ridge2 <- function(X, start, lambda) {
	return (lambda_crossval_objective2(X, start, lambda1 = 0.0, lambda))
}

lambda_opt_ridge <- function(X, start) {
	lambda_0 <- optimize(lambda_crossval_ridge, X = X, start = start, c(0.0, 0.5))
	lambda_min <- lambda_0$minimum
	
	zero_min = lambda_crossval_ridge(X, start, 0.0)
	
	if(zero_min < lambda_0$objective) {
		return (0.0)
	}
	
	return (lambda_min)
}

lambda_opt_ridge2 <- function(X, start) {
	lambda_0 <- optimize(lambda_crossval_ridge2, X = X, start = start, c(0.0, 0.5))
	lambda_min <- lambda_0$minimum
	
	zero_min = lambda_crossval_ridge2(X, start, 0.0)
	
	if(zero_min < lambda_0$objective) {
		return (0.0)
	}
	
	return (lambda_min)
}

compute_bucketset <- function(theta) {
	mu <- theta[1]
	sigma <- theta[2]
	
	lower <- mu - (10 * sigma);
	upper <- mu + (10 * sigma);
	intervalSize <- upper - lower
	bucketSet <- lower + (0:floor(intervalSize * buckets) / buckets)
	
	return (bucketSet)
}


rho <- function(theta0, theta) {
	bucketSet <- compute_bucketset(theta0)
	
	p <- probCalculation(bucketSet, theta0)
	q <- probCalculation(bucketSet, theta)
	
	ratio <- pmax(p / q, 1.0e-20) # floor this out, lim p -> 0 of p log(p) = 0. 
	kl <- sum(p * log(ratio))
	return (kl)
}

hf <- function(theta0) {
	bucketSet <- compute_bucketset(theta0)	
	p <- pmax(probCalculation(bucketSet, theta0), 1.0e-20)
	hval <- -sum(p*log(p))
	return (hval)
}


run_test <- function(mu, sigma, startTheta, n, iteration) {
	#Generate test data.
	# LDDP this thing with 5 sigma cutoffs. 
	# snap mean and sigma to grid. 
	mu <- snapToGrid(mu, buckets)
	sigma <- max(snapToGrid(sigma, buckets), 0.1)
	
	bucketSet <- compute_bucketset(c(mu, sigma))
	#print(paste("Bucket sum: ", sum(probCalculation(bucketSet, c(mu, sigma))))) # Should be extremely close to 1.0
	
	X <- (sigma * rnorm(n)) + mu
	X <- snapToGrid(X, buckets)
	
	np = 20 * 1000
	Xp <- (sigma * rnorm(np)) + mu
	Xp <- snapToGrid(Xp, buckets)
	
	mle <- solve_mle(X, startTheta, 0.0, 0.0)
	mle_entropy <- objective_mle(X, mle, 0.0, 0.0)

	lambda2 = lambda_opt_ridge(X, startTheta)
	l2_theta = solve_mle(X, startTheta, 0.0, lambda2)
	l2_entropy <- objective_mle(X, l2_theta, 0.0, 0.0)

	lambda2a= lambda_opt_ridge2(X, startTheta)
	l2a_theta = solve_mle2(X, startTheta, 0.0, lambda2a)
	l2a_entropy <- objective_mle(X, l2a_theta, 0.0, 0.0)

	ice_theta = solve_ice(X, X, startTheta)
	ice_theta2 = solve_ice(X, Xp, startTheta)
	ice_entropy <- objective_mle(X, ice_theta, 0.0, 0.0)
	true_entropy <- objective_mle(X, c(mu, sigma), 0.0, 0.0)
	hf <- hf(c(mu, sigma))

	mle_kl = rho(c(mu, sigma), mle)
	l2_kl = rho(c(mu, sigma), l2_theta)
	l2a_kl = rho(c(mu, sigma), l2a_theta)
	ice_kl = rho(c(mu, sigma), ice_theta)
	ice_kl2 = rho(c(mu, sigma), ice_theta2)

	#print(paste("KL Diff: ", ice_kl, " -> ", ice_kl2))

	iMatrix = ice_iMatrix(X, ice_theta)
	jMatrix = ice_jMatrix(X, ice_theta)

	iMatrixT = ice_iMatrix(Xp, ice_theta)
	jMatrixT = ice_jMatrix(Xp, ice_theta)

	min_condition = min(rcond(iMatrixT), rcond(jMatrixT))
	min_conditionTrue = min(rcond(iMatrix), rcond(jMatrix))
	
	a = pinv(jMatrix)
	b = pinv(diag(2)*jMatrix)
	c = pinv(jMatrixT)
	
	iceMatrix1 = iMatrix %*% a
	iceMatrix2 = iMatrix %*% b
	iceMatrix3 = iMatrixT %*% c
	iceCorrection1 = max(sum(diag(iceMatrix1)), 0) / length(X)
	iceCorrection2 = max(sum(diag(iceMatrix2)), 0) / length(X)
	iceCorrection3 = max(sum(diag(iceMatrix3)), 0) / length(X)

	norm1 = norm(a)
	norm2 = norm(b)
	norm3 = norm(a - b)	
	norm4 = norm(c)
	norm5 = norm(c - a)
	norm6 = norm(c - b)

	#print(paste("MLE: ", mle_entropy, " -> ", mle_kl))
	#print(paste("L2[", lambda2, "]: ", l2_entropy, " -> ", l2_kl))
	
	print(paste(n, ", ", iteration, ", ", mu, ", ", sigma, ", ", mle[1], ", ", mle[2], ", ", lambda2, ", ", l2_theta[1], ", ", l2_theta[2], ", ", l2a_theta[1], ", ", l2a_theta[2], ", ", ice_theta[1], ", ", ice_theta[2], ", ", mle_kl, ", ", l2_kl, ", ", l2a_kl, ", ", ice_kl, ", ", mle_entropy, ", ", l2_entropy, ", ", l2a_entropy, ", ", ice_entropy, ", ", true_entropy, ", ", hf, ", ", min_condition, ", ", min_conditionTrue, ", ", iceCorrection1, ", ", iceCorrection2, ", ", iceCorrection3, ", ", norm1, ", ", norm2, ", ", norm3, ", ", norm4, ", ", norm5, ", ", norm6, ", ", ice_kl2))
	
	
	return (c(mle_kl, l2_kl, l2a_kl, ice_kl, ice_theta[1], ice_theta[2], ice_theta[1] - mle[1], ice_theta[2] - mle[2]))
}

run_test_repeated <- function(mu, sigma, startTheta, n, count) {
	results <- c()

	print("n, iteration, actualMu, actualSigma, mleMu, mleSigma, l2Lambda, l2Mu, l2Sigma, l2aMu, l2aSigma, iceMu, iceSigma, mleKl, l2Kl, l2aKl, iceKl, mleTrainEntropy, l2TrainEntropy, l2aTrainEntropy, iceTrainEntropy, trueTrainEntropy, hf, min_condition, min_conditionTrue, jCorrecetion, dCorrection, trueCorrection, jNorm, dNorm, diffNorm, trueNorm, trueApproxDiffNorm, trueVsDiagNorm, iceKlTrueJ")
	
	for(k in 1:count) {
		output <- run_test(mu, sigma, startTheta, n, k)
		results <- cbind(c(output), results)
	}
	
	repeated_results <- colMeans(t(results))
	repeated_devs <- sqrt(colVars(t(results)) / n)
	l2_diffs = (t(results[2,]) - t(results[1,]))[1,]
	ice_diffs = (t(results[4,]) - t(results[1,]))[1,]
	
	l2_dev = sqrt(var(l2_diffs) / count)
	ice_dev = sqrt(var(ice_diffs) / count)
	
	l2_z = mean(l2_diffs) / l2_dev;
	ice_z = mean(ice_diffs) / ice_dev;
	
	ice_theta_mu = t(results[5,])[1,];
	ice_theta_dev = t(results[6,])[1,];
	ice_diff_mu = t(results[7,])[1,];
	ice_diff_dev = t(results[8,])[1,];
	
	ice_var_mu = sqrt(var(ice_diff_mu) / count)
	ice_var_dev = sqrt(var(ice_diff_dev) / count)
	
	ice_mu_z = mean(ice_diff_mu) / ice_var_mu;
	ice_dev_z = mean(ice_diff_dev) / ice_var_dev;
	
	
	print(paste("MLE KL: ", repeated_results[1]))
	print(paste("L2 KL: ", repeated_results[2], " (worse by ", (repeated_results[2] - repeated_results[1]), ")"))
	print(paste("L2a KL: ", repeated_results[3], " (worse by ", (repeated_results[3] - repeated_results[1]), ")"))
	print(paste("ICE KL: ", repeated_results[4], " (worse (or better if negative) by ", (repeated_results[4] - repeated_results[1]), ") [", ice_z , "]"))
	
	
	print(paste("Summary: ", repeated_results[1], ", ", repeated_devs[1], ", ", repeated_results[2], ", ", repeated_devs[2], ", ", repeated_results[3], ", ", repeated_devs[3], ", ", repeated_results[4], ", ", repeated_devs[4], ", ", l2_z, ", ", ice_z, ", [", mean(ice_diff_mu), ", ", mean(ice_diff_dev), " : ", ice_mu_z, ", ", ice_dev_z, "]"))
	
	return (repeated_results)
}


meanTest = rnorm(1)[1]
sigmaTest = abs(rnorm(1)[1])

start <- c(1.0, 1.0)

run_test(0.0, 0.2, start, 10, 0)


# This should almost always show that L2 is worse than MLE. 
# repeated_results <- run_test_repeated(0.2, 0.2, start, 4, 500)
# repeated_results <- run_test_repeated(0.2, 0.2, start, 8, 500)
repeated_results <- run_test_repeated(0.2, 0.2, start, 16, 2000)
repeated_results <- run_test_repeated(0.2, 0.2, start, 32, 2000)
repeated_results <- run_test_repeated(0.2, 0.2, start, 64, 2000)
repeated_results <- run_test_repeated(0.2, 0.2, start, 128, 2000)
repeated_results <- run_test_repeated(0.2, 0.2, start, 256, 2000)
repeated_results <- run_test_repeated(0.2, 0.2, start, 512, 2000)
repeated_results <- run_test_repeated(0.2, 0.2, start, 1024, 2000)

# print(paste("MLE KL: ", repeated_results[1]))
# print(paste("L2 KL: ", repeated_results[2], " (worse by ", (repeated_results[2] - repeated_results[1]), ")"))
# print(paste("L2a KL: ", repeated_results[3], " (worse by ", (repeated_results[3] - repeated_results[1]), ")"))
# print(paste("ICE KL: ", repeated_results[4], " (worse (or better if negative) by ", (repeated_results[4] - repeated_results[1]), ")"))

